/* vim: set ai ts=8 sw=4 sts=4 noet: */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <vorbis/codec.h>
#include <vorbis/vorbisfile.h>
#include <id3tag.h>
#include "liteamp.h"
#include "la-track.h"
#include "util.h"

/* private------------------------------------------------*/

static GSList *fallback_codesets = NULL;
static gchar *used_codeset = NULL;

//
// id3v1 supports
//

static gchar *get_id3_tag_string(struct id3_tag *tag, char *id)
{
    struct id3_frame *frame;
    gboolean is_genre_frame;
    GString *str;
    union id3_field *field;
    gint nstrings, index;
    id3_ucs4_t const *ucs4;
    gchar *latin1;
    gchar *utf8;

    frame = id3_tag_findframe(tag, id, 0);
    if (frame) {
	is_genre_frame = !strcmp(id, ID3_FRAME_GENRE);

	str = g_string_sized_new(32);

	field = &frame->fields[1];
	nstrings = id3_field_getnstrings(field);

	for (index = 0; index < nstrings; index++) {
	    ucs4 = id3_field_getstrings(field, index);
	    if (!ucs4)
		return NULL;

	    if (is_genre_frame)
		ucs4 = id3_genre_name(ucs4);

	    // some troubles with utf8duplicate :(
	    // so, i'd used latin1 and convert it utf8 by hand
	    //utf8 = id3_ucs4_utf81duplicate(ucs4);
	    //if (!utf8) break;
	    //g_string_append(str, utf8);

	    latin1 = id3_ucs4_latin1duplicate(ucs4);
	    if (!latin1)
		break;
	    g_string_append(str, latin1);
	}

	//utf8 = str->str;
	//g_string_free(str, FALSE);
	//return utf8;

	utf8 =
	    la_str_to_utf8_with_fallback(str->str,
					 fallback_codesets, &used_codeset);
	g_string_free(str, FALSE);
	return utf8;
    }

    return NULL;
}

static gchar *get_id3_tag_comment(struct id3_tag *tag)
{
    struct id3_frame *frame;
    id3_ucs4_t const *ucs4;
    gint index;
    gchar *latin1;
    gchar *utf8;

    index = 0;
    while ((frame = id3_tag_findframe(tag, ID3_FRAME_COMMENT, index++))) {
	id3_latin1_t *ptr, *newline;
	int first = 1;

	ucs4 = id3_field_getstring(&frame->fields[2]);
	if (!ucs4)
	    return NULL;

	if (*ucs4)
	    continue;

	ucs4 = id3_field_getfullstring(&frame->fields[3]);
	if (!ucs4)
	    return NULL;

	latin1 = id3_ucs4_latin1duplicate(ucs4);

	utf8 =
	    la_str_to_utf8_with_fallback(latin1, fallback_codesets,
					 &used_codeset);
	g_free(latin1);
	return utf8;
    }

    return NULL;
}

static gboolean read_local_mp3(LaTrack * track, const gchar * filename)
{
    struct id3_file *file;
    struct id3_tag *tag;
    struct id3_frame *frame;
    id3_ucs4_t *ucs4;
    gchar *utf8;

    gint seconds;
    gint frames;

    la_track_set_channels(track, 2);
    la_track_set_samplerate(track, 0);
    la_track_set_bitrate(track, 0);

    if ((decoder_get_total_time_frame(filename, &seconds, &frames)) == -1) {
	la_track_set_seconds(track, 0);
	la_track_set_frames(track, 0);
	g_message("mp3 file corruped:%s\n", filename);
	return FALSE;
    }
    la_track_set_seconds(track, seconds);
    la_track_set_frames(track, frames);

    file = id3_file_open(filename, ID3_FILE_MODE_READONLY);
    if (!file)
	return FALSE;

    tag = id3_file_tag(file);
    if (!tag)
	return FALSE;

    utf8 = get_id3_tag_string(tag, ID3_FRAME_TITLE);
    if (utf8) {
	la_track_set_title(track, utf8);
	free(utf8);
    }
    utf8 = get_id3_tag_string(tag, ID3_FRAME_ARTIST);
    if (utf8) {
	la_track_set_artist(track, utf8);
	free(utf8);
    }
    utf8 = get_id3_tag_string(tag, ID3_FRAME_ALBUM);
    if (utf8) {
	la_track_set_album(track, utf8);
	free(utf8);
    }
    utf8 = get_id3_tag_string(tag, ID3_FRAME_TRACK);
    if (utf8) {
	la_track_set_track(track, utf8);
	free(utf8);
    }
    utf8 = get_id3_tag_string(tag, ID3_FRAME_GENRE);
    if (utf8) {
	la_track_set_genre(track, utf8);
	free(utf8);
    }
    utf8 = get_id3_tag_comment(tag);
    if (utf8) {
	la_track_set_comment(track, utf8);
	free(utf8);
    }

    id3_tag_delete(tag);
    id3_file_close(file);

    return TRUE;
}

//
// ogg-vorbis supports
//

static gboolean read_local_ogg(LaTrack * track, const gchar * filename)
{
    FILE *fp;
    OggVorbis_File vf;
    vorbis_info *vi;
    GString *comment_str;
    char **ptr;
    gchar **tokens;
    gchar *utf8;

    if ((fp = fopen(filename, "rb")) == NULL)
	return FALSE;

    if (ov_open(fp, &vf, NULL, 0) < 0)
	return FALSE;

    vi = ov_info(&vf, -1);
    la_track_set_bitrate(track, vi->bitrate_nominal);
    la_track_set_samplerate(track, vi->rate);
    la_track_set_frames(track, ov_pcm_total(&vf, -1));
    la_track_set_channels(track, vi->channels);
    la_track_set_seconds(track, ov_time_total(&vf, -1));

    comment_str = g_string_sized_new(32);

    ptr = ov_comment(&vf, -1)->user_comments;
    while (*ptr) {
	tokens = g_strsplit(*ptr, "=", 2);

	if (LA_STR_IS_NOT_EMPTY(tokens[1])) {
	    utf8 =
		la_str_to_utf8_with_fallback(tokens[1],
					     fallback_codesets,
					     &used_codeset);
	    g_ascii_strup(tokens[0], -1);
	    if (!strcmp(tokens[0], "TITLE")) {

		la_track_set_title(track, utf8);

	    } else if (!strcmp(tokens[0], "ARTIST")) {

		la_track_set_artist(track, utf8);

	    } else if (!strcmp(tokens[0], "ALBUM")) {

		la_track_set_album(track, utf8);

	    } else if (!strcmp(tokens[0], "TRACK")
		       || !strcmp(tokens[0], "TRACKNUMBER")) {

		la_track_set_track(track, utf8);

	    } else if (!strcmp(tokens[0], "YEAR")) {

		la_track_set_year(track, utf8);

	    } else if (!strcmp(tokens[0], "GENRE")) {

		la_track_set_genre(track, utf8);
	    }

	    g_free(utf8);
	} else {
	    g_string_append(comment_str, *ptr);
	    g_string_append_c(comment_str, '\n');
	}
	g_strfreev(tokens);

	++ptr;
    }
    ov_clear(&vf);

    if (LA_STR_IS_NOT_EMPTY(comment_str->str)) {
	utf8 =
	    la_str_to_utf8_with_fallback(comment_str->str,
					 fallback_codesets, &used_codeset);
	la_track_set_comment(track, utf8);
    }

    g_string_free(comment_str, TRUE);

    return TRUE;
}

/* public-------------------------------------------------*/

void tag_init(GSList * _fallback_codesets)
{
    fallback_codesets = _fallback_codesets;
}

gboolean tag_read(LaTrack * track)
{
    GnomeVFSURI *uri;
    GnomeVFSFileInfo *info;
    const gchar *mime_type;
    gchar *local_path;
    gchar *pseudo_title;

    uri = la_track_get_vfs_uri(track);
    info = la_track_get_vfs_info(track);

    // get physical file info
    la_track_set_bytes(track, info->size);
    la_track_set_timestamp(track, info->mtime);

    if (GNOME_VFS_FILE_INFO_LOCAL(info)) {
	mime_type = gnome_vfs_file_info_get_mime_type(info);

	local_path = la_uri_to_local(la_track_get_uri(track));

	if (!strcmp(MIME_TYPE_OGG, mime_type)) {
	    // read ogg-voribs tag
	    read_local_ogg(track, local_path);
	} else if (!strcmp(MIME_TYPE_MP3, mime_type)) {
	    // read id3v1/v2 tag
	    read_local_mp3(track, local_path);
	} else {
	    //TODO: some fallback parsing needed. what's that?
	    g_message("unknown file type %s for %s\n", mime_type,
		      local_path);
	}
	g_free(local_path);
    } else {
	//TODO: what can i do for remote file?
    }

    // provides fallback title from uri
    if (LA_STR_IS_EMPTY(la_track_get_title(track))) {
	pseudo_title = gnome_vfs_uri_extract_short_name(uri);
	la_track_set_title(track, pseudo_title);
	g_free(pseudo_title);
    }

    gnome_vfs_file_info_unref(info);
    gnome_vfs_uri_unref(uri);

    return TRUE;
}

gboolean tag_write(LaTrack * track)
{
    //TODO:
    //not yet implemented
    return FALSE;
}

gboolean tag_remove(LaTrack * track)
{
    //TODO:
    //not yet implemented
    return FALSE;
}

/* la-tag.c */
